#!/bin/sh
#
# Copyright (c) 2014, 2016, The Linux Foundation. All rights reserved.
#
. /lib/sn_functions.sh
. /lib/wifi/wifi_funcs.sh

append DRIVERS "qcawifi"
default_cfg80211=1

wifi0_mac_file="/tmp/wifi0_mac.txt"
wifi1_mac_file="/tmp/wifi1_mac.txt"
wifi2_mac_file="/tmp/wifi2_mac.txt"
error_msg="/tmp/invalid_mac.txt"

ar71xx_sh_file="/lib/ar71xx.sh"
ipq40xx_sh_file="/lib/firmware/IPQ4019"
ipq806x_sh_file="/lib/ipq806x.sh"

platform_name="unknown"

if [ -e $ar71xx_sh_file ]; then
	platform_name="ar71xx"
elif [ -e $ipq40xx_sh_file ]; then
	platform_name="ipq40xx"
elif [ -e $ipq806x_sh_file ]; then
	platform_name="ipq806x"
fi 

monitor_mode() {	#non-using function now, monitor mode runs normal flow to create, keep it for reference
	local device="$1"

	config_get opmode "$device" opmode
	[ "$opmode" == "mon" ] && {
		config_get vifs "$device" vifs
		config_get phy "$device" phy
		for vif in $vifs; do
			config_get ifname "$vif" ifname
			config_get mode_display "$vif" mode_display
			[ "$opmode" == "$mode_display" ] && {
				DebugPrint "Create monitor vap: wlanconfig $ifname create wlandev $phy wlanmode $opmode"
				ifname=$(/usr/sbin/wlanconfig "$ifname" create wlandev "$phy" wlanmode "$opmode")
				[ $? -ne 0 ] && DebugPrint "monitor_mode($device): Failed to set up $mode vif $ifname" || ifconfig "$ifname" up
				return 0
			}
		done
	}
	return 1
}

find_qcawifi_phy() {
	local device="$1"

	local macaddr="$(config_get "$device" macaddr | tr 'A-Z' 'a-z')"
	config_get phy "$device" phy
	[ -z "$phy" -a -n "$macaddr" ] && {
		cd /sys/class/net
		for phy in $(ls -d wifi* 2>&-); do
			[ "$macaddr" = "$(cat /sys/class/net/${phy}/address)" ] || continue
			config_set "$device" phy "$phy"
			break
		done
		config_get phy "$device" phy
	}
	[ -n "$phy" -a -d "/sys/class/net/$phy" ] || {
		DebugPrint "phy for wifi device $1 not found"
		return 1
	}
	[ -z "$macaddr" ] && {
		config_set "$device" macaddr "$(cat /sys/class/net/${phy}/address)"
	}
	return 0
}
do_cold_boot_calibration() {
	local board_name
	[ -f /tmp/sysinfo/board_name ] && {
		board_name=$(cat /tmp/sysinfo/board_name)
	}

	case "$board_name" in
	ap-hk01-c1|ap-hk02)
		daemon=`cat /sys/module/cnss2/parameters/daemon_support`
		[ $daemon = "N" ] && echo "******No daemon support*****"  > /dev/console && return 0

		[ -f /tmp/cold_boot_done ] && echo "******Not the first boot. Skip coldboot calibration*****"  > /dev/console && return 0

		echo "*********Starting cnssdaemon and initiating cold boot calibration*************" > /dev/console
		mkdir -p /data/misc/wifi/
		/usr/bin/cnssdaemon > /dev/console 2>&1 &
		insmod qca_ol testmode=7
		# insmod and rmmod wifi_3_0 kernel object in addition to qca_ol
		# as wifi_3_0 initiates and shuts down driver.
		insmod wifi_3_0
		rmmod wifi_3_0
		rmmod qca_ol
		sync
		touch /tmp/cold_boot_done
		;;

	*)
		echo "INFO: Cold boot calibration not applicable: $board_name"  > /dev/console
		;;
	esac
}

scan_qcawifi() {
	local device="$1"
	local wds
	local adhoc sta ap disabled

	[ ${device%[0-9]} = "wifi" ] && config_set "$device" phy "$device"

	local ifidx=0
	local radioidx=${device#wifi}

	config_get vifs "$device" vifs
	config_get backgroundscanEnable "$device" backgroundscanEnable
	config_get apsteer "$device" apsteer 0
	config_get spectrumMode "$device" spectrumMode 0
	config_get opmode "$device" opmode
	haveScanRadio=$(check_scan_radio_enable)

	[ "$spectrumMode" == "1" ] && isCreat=0

	for vif in $vifs; do
		config_get mode "$vif" mode 

		[ "$spectrumMode" == "1" -a "$opmode" == "ap" -a "$haveScanRadio" == "0" ] && {
			config_get mode_display "$vif" mode_display
			[ "$mode_display" == "ap" ] && {
				[ "$isCreat" == "0" ] && {
					config_set "$vif" disabled 0
					uci set wireless.$vif.disabled="0"
					isCreat=1
				} || {
					config_set "$vif" disabled 1
					uci set wireless.$vif.disabled="1"
				}
				uci commit wireless
			}
		}

		[ "$haveScanRadio" == "0" -a "$spectrumMode" == "0" ] && [ "$mode" == "monitor" ] && {
			config_get disabled "$vif" disabled
			if [ "$backgroundscanEnable" == "1" -o "$apsteer" == "1" ]; then
				if [ "$(uci -q get wifiprofile.snWifiConf.SUPPORT_CAPWAP_IWLIST_SCAN)" == "1" ]; then
					[ "$disabled" == "0" ] && {
						config_set "$vif" disabled 1
						uci set wireless.$vif.disabled="1"
						uci commit wireless
						DebugPrint "force to disable monitor mode because background scanning replace to iwlist displayscan."
					}
				else
					[ "$disabled" == "1" ] && {
						config_set "$vif" disabled 0
						uci set wireless.$vif.disabled="0"
						uci commit wireless
						DebugPrint "force to enable monitor mode because background scanning need airodump-ng."
					}
				fi
			else
				[ "$disabled" == "0" ] && {
					config_set "$vif" disabled 1
					uci set wireless.$vif.disabled="1"
					uci commit wireless
					DebugPrint "force to disable monitor mode because the background scanning has been terminated."
				}
			fi
		}
	
		config_get_bool disabled "$vif" disabled 0

		# senao WAR : mesh enable -> SCAN SSID disable
		[ "$(uci -q get wifiprofile.snWifiConf.SUPPORT_WAR_MESH_DISABLE_SCAN_SSID)" == "1" -a "$mode" == "monitor" -a "$(uci get mesh.wifi.disabled)" == "0" ] && disabled=1

		# senao WAR : mpsk enable -> SCAN SSID and Background disable
		mpsk_enable=0
		mpsk_ifs=$(foreach wireless wifi-iface multi_group_key 1)
		for mpsk_if in $mpsk_ifs; do
			[ "$(uci get wireless.$mpsk_if.disabled)" = "0" ] && mpsk_enable=1
		done
		[ "$(uci -q get wifiprofile.snWifiConf.SUPPORT_WAR_MPSK_DISABLE_BACKGROUNDSCAN)" == "1" -a $mpsk_enable -eq 1 -a "$mode" == "monitor" ] && disabled=1

		[ $disabled = 0 ] || continue

	#	local vifname
	#	[ $ifidx -gt 0 ] && vifname="ath${radioidx}$ifidx" || vifname="ath${radioidx}"

		config_get ifname "$vif" ifname
	#	config_set "$vif" ifname "${ifname:-$vifname}"

		case "$mode" in
			adhoc|sta|ap|monitor|wrap)
				append $mode "$vif"
			;;
			wds)
				config_get ssid "$vif" ssid
				[ -z "$ssid" ] && continue

				config_set "$vif" wds 1
				config_set "$vif" mode sta
				mode="sta"
				addr="$ssid"
				${addr:+append $mode "$vif"}
			;;
			*) DebugPrint "$device($vif): Invalid mode, ignored."; continue;;
		esac

		ifidx=$(($ifidx + 1))
	done

	case "${adhoc:+1}:${sta:+1}:${ap:+1}" in
		# valid mode combinations
		1::) wds="";;
		1::1);;
		:1:1)config_set "$device" nosbeacon 1;; # AP+STA, can't use beacon timers for STA
		:1:);;
		::1);;
		::);;
		*) DebugPrint "$device: Invalid mode combination in config"; return 1;;
	esac
	config_get hwmode $device hwmode
	config_get channel $device channel

	#WAR, jacky: monitor vap makes 11ac auto channel state-machine abnormal, and then wpa_cli cannot add config properly, so we create monitor before ap mode vap
	if [ "$hwmode" == "11ac" -a "$channel" == "auto" ]; then
		config_set "$device" vifs "${monitor:+$monitor }${sta:+$sta }${ap:+$ap }${wrap:+$wrap }${adhoc:+$adhoc }${wds:+$wds}"
	else
		config_set "$device" vifs "${sta:+$sta }${ap:+$ap }${wrap:+$wrap }${adhoc:+$adhoc }${wds:+$wds }${monitor:+$monitor}"
	fi
}

# The country ID is set at the radio level. When the driver attaches the radio,
# it sets the default country ID to 840 (US STA). This is because the desired
# VAP modes are not known at radio attach time, and STA functionality is the
# common unit of 802.11 operation.
# If the user desires any of the VAPs to be in AP mode, then we set a new
# default of 843 (US AP with TDWR) from this script. Even if any of the other
# VAPs are in non-AP modes like STA or Monitor, the stricter default of 843
# will apply.
# No action is required here if none of the VAPs are in AP mode.
set_default_country() {
	local device="$1"
	local mode

	find_qcawifi_phy "$device" || return 1
	config_get phy "$device" phy

        # If the country parameter is number (either hex or decimal), we
        # assume it's a regulatory domain - i.e. we use device_if setCountryID.
        # Else we assume it's a country code - i.e. we use device_if setCountry.
        country=$(uci get wireless.$device.country)
        if [ `expr "$country" : '[0-9].*'` -ne 0 ]; then
                if [ "$country" = "00" -a "$domain" = "2" ]; then
                    device_if "$phy" setCountryID 528
                else
                    device_if "$phy" setCountryID "$country"
                fi

        elif [ -n "$country" ]; then
                device_if "$phy" setCountry "$country"
        else
	# The country ID is set at the radio level. When the driver attaches the radio,
	# it sets the default country ID to 840 (US STA). This is because the desired
	# VAP modes are not known at radio attach time, and STA functionality is the
	# common unit of 802.11 operation.
	# If the user desires any of the VAPs to be in AP mode, then we set a new
	# default of 840 (US AP with TDWR) from this script. Even if any of the other
	# VAPs are in non-AP modes like STA or Monitor, the stricter default of 843
	# will apply.
	# No action is required here if none of the VAPs are in AP mode.
		config_get vifs "$device" vifs
		for vif in $vifs; do
			config_get_bool disabled "$vif" disabled 0
			[ $disabled = 0 ] || continue

			config_get mode "$vif" mode
			case "$mode" in
				ap|wrap|ap_monitor|ap_smart_monitor|ap_lp_iot)
					device_if "$phy" setCountryID 840
					return 0;
				;;
			*) ;;
			esac
		done
	fi
	return 0
}

config_low_targ_clkspeed() { #non-using function now
        local board_name
        [ -f /tmp/sysinfo/board_name ] && {
                board_name=$(cat /tmp/sysinfo/board_name)
        }

        case "$board_name" in
                ap147 | ap151)
                   echo "true"
                ;;
                *) echo "false"
                ;;
        esac
}

# configure tx queue fc_buf_max
config_tx_fc_buf() {
	local phy="$1"
	local board_name
	[ -f /tmp/sysinfo/board_name ] && {
		board_name=$(cat /tmp/sysinfo/board_name)
	}
	memtotal=$(grep MemTotal /proc/meminfo | awk '{print $2}')

	case "$board_name" in
		ap-dk*)
			if [ $memtotal -le 131072 ]; then
				# 4MB tx queue max buffer size
				device_if "$phy" fc_buf_max 4096
				device_if "$phy" fc_q_max 512
				device_if "$phy" fc_q_min 32
			elif [ $memtotal -le 256000 ]; then
				# 8MB tx queue max buffer size
				device_if "$phy" fc_buf_max 8192
				device_if "$phy" fc_q_max 1024
				device_if "$phy" fc_q_min 64
			fi
				# default value from code memsize > 256MB
		;;

		*)
		;;
	esac
}

load_qcawifi() {

	local qca_da_needed=0
	local qca_ol_needed=0
	local reload=0
	local hk_ol_num=0

	echo -n "/ini" > /sys/module/firmware_class/parameters/path
	wifiProfile2Config	#jacky prepare wifiprofile config before kmod insmod, move from enable_qcawifi here for detect_qcawifi maybe execute before enable_qcawifi

        is_wal=`grep waltest_mode /proc/cmdline | wc -l`
        [ $is_wal = 1 ] && waltest_qcawifi && return

	if [ ! -d "/cfg/default" ]; then
		mkdir -p /cfg/default
		cp -rf /ini/* /cfg/default/
		sync #senao patch for file protection
	fi

	cp /cfg/default/* /ini/
	sync #senao patch for file protection

        local enable_cfg80211=`uci show qcacfg80211.config.enable |grep "qcacfg80211.config.enable='1'"`
        [ -n "$enable_cfg80211" ] && echo "qcawifi configuration is disable" > /dev/console && return 1;

        lock /var/run/wifilock
        [ -f /tmp/sysinfo/board_name ] && {
                board_name=$(cat /tmp/sysinfo/board_name)
        }
        memtotal=$(grep MemTotal /proc/meminfo | awk '{print $2}')

        case "$board_name" in
                ap-dk01.1-c1 | ap-dk01.1-c2 | ap-dk04.1-c1 | ap-dk04.1-c2 | ap-dk04.1-c3)
                        if [ $memtotal -le 131072 ]; then
                                echo 1 > /proc/net/skb_recycler/max_skbs
                                echo 1 > /proc/net/skb_recycler/max_spare_skbs
                                append umac_args "low_mem_system=1"
                        fi
                ;;
                ap152 | ap147 | ap151 | ap135 | ap137)
                        if [ $memtotal -le 66560 ]; then
                                def_pktlog_support=0
                        fi
                ;;
        esac

	set_global_ini	#do it after checked cfg/default

        config_get_bool testmode qcawifi testmode
        [ -n "$testmode" ] && append ol_args "testmode=$testmode"

        config_get_bool ar900b_emu qcawifi ar900b_emu
        [ -n "$ar900b_emu" ] && append ol_args "ar900b_emu=$ar900b_emu"

        config_get frac qcawifi frac
        [ -n "$frac" ] && append ol_args "frac=$frac"

        config_get intval qcawifi intval
        [ -n "$intval" ] && append ol_args "intval=$intval"

        #config_get atf_mode qcawifi atf_mode
        #[ -n "$atf_mode" ] && append umac_args "atf_mode=$atf_mode"

        config_get atf_msdu_desc qcawifi atf_msdu_desc
        [ -n "$atf_msdu_desc" ] && append umac_args "atf_msdu_desc=$atf_msdu_desc"

        config_get atf_peers qcawifi atf_peers
        [ -n "$atf_peers" ] && append umac_args "atf_peers=$atf_peers"

        config_get debugtxbf_ulofdma qcawifi debugtxbf_ulofdma
        [ "$debugtxbf_ulofdma" = "Y" ] && echo Y > /sys/module/cnss2/parameters/qca6290_support
        [ "$debugtxbf_ulofdma" = "N" ] && echo N > /sys/module/cnss2/parameters/qca6290_support

        config_get atf_max_vdevs qcawifi atf_max_vdevs
        [ -n "$atf_max_vdevs" ] && append umac_args "atf_max_vdevs=$atf_max_vdevs"

        config_get ar900b_20_targ_clk qcawifi ar900b_20_targ_clk
        [ -n "$ar900b_20_targ_clk" ] && append ol_args "ar900b_20_targ_clk=$ar900b_20_targ_clk"

        config_get qca9888_20_targ_clk qcawifi qca9888_20_targ_clk
        [ -n "$qca9888_20_targ_clk" ] && append ol_args "qca9888_20_targ_clk=$qca9888_20_targ_clk"

        cfg_low_targ_clkspeed=$(config_low_targ_clkspeed)
        [ -z "$qca9888_20_targ_clk" ] && [ $cfg_low_targ_clkspeed = "true" ] && append ol_args "qca9888_20_targ_clk=300000000"

        config_get prealloc_disabled qcawifi prealloc_disabled
        [ -n "$prealloc_disabled" ] && append qdf_args "prealloc_disabled=$prealloc_disabled"

        if [ -n "$nss_wifi_olcfg" ] && [ "$nss_wifi_olcfg" != "0" ]; then
		local mp_256="$(ls /proc/device-tree/ | grep -rw "MP_256")"
		local mp_512="$(ls /proc/device-tree/ | grep -rw "MP_512")"
                sysctl dev.nss.n2hcfg.n2h_high_water_core0 >/dev/null 2>/dev/null
		
		#update the ini nss info
		update_ini_nss_info

		#If this is a first time load, then remove the one radio up file
		if [ ! -d /sys/module/qca_ol ] && [ -f /tmp/wifi_nss_up_one_radio ]; then
			rm /tmp/wifi_nss_up_one_radio
		fi
	
		if [ "$mp_256" == "MP_256" ]; then
			#total pbuf size is 160 bytes,allocate memory for 8712 pbufs
			sysctl -w dev.nss.n2hcfg.extra_pbuf_core0=1400000 >/dev/null 2>/dev/null
			sysctl -w dev.nss.n2hcfg.n2h_high_water_core0=20432 >/dev/null 2>/dev/null
			sysctl -w dev.nss.n2hcfg.n2h_wifi_pool_buf=0 >/dev/null 2>/dev/null
		elif [ "$mp_512" == "MP_512" ]; then
			[ -d /sys/module/qca_ol ] || { \
			hk_ol_num="$(cat /lib/wifi/wifi_nss_hk_olnum)"
			if [ $hk_ol_num -eq 3 ]; then
				#total pbuf size is 160 bytes,allocate memory for 19928 pbufs
				sysctl -w dev.nss.n2hcfg.extra_pbuf_core0=3200000 >/dev/null 2>/dev/null
				sysctl -w dev.nss.n2hcfg.n2h_high_water_core0=31648 >/dev/null 2>/dev/null
				sysctl -w dev.nss.n2hcfg.n2h_wifi_pool_buf=0 >/dev/null 2>/dev/null
			else
				#total pbuf size is 160 bytes,allocate memory for 18904 pbufs
				sysctl -w dev.nss.n2hcfg.extra_pbuf_core0=3100000 >/dev/null 2>/dev/null
				sysctl -w dev.nss.n2hcfg.n2h_high_water_core0=30624 >/dev/null 2>/dev/null
				sysctl -w dev.nss.n2hcfg.n2h_wifi_pool_buf=8192 >/dev/null 2>/dev/null
			fi
		}
		else
		        case "$board_name" in
			ap-hk09*)
				local soc_version_major
				[ -f /sys/firmware/devicetree/base/soc_version_major ] && {
					soc_version_major="$(hexdump -n 1 -e '"%1d"' /sys/firmware/devicetree/base/soc_version_major)"
				}
				if [ $soc_version_major = 2 ];then
					[ -d /sys/module/qca_ol ] || { \
						#total pbuf size is 160 bytes,allocate memory for 55672 pbufs
						sysctl -w dev.nss.n2hcfg.extra_pbuf_core0=9000000 >/dev/null 2>/dev/null
						sysctl -w dev.nss.n2hcfg.n2h_high_water_core0=67392 >/dev/null 2>/dev/null
						#initially after init 4k buf for 5G and 4k for 2G will be allocated, later range will be configured
						sysctl -w dev.nss.n2hcfg.n2h_wifi_pool_buf=40960 >/dev/null 2>/dev/null
					}
				else
					#total pbuf size is 160 bytes,allocate memory for 57184 pbufs
					sysctl -w dev.nss.n2hcfg.extra_pbuf_core0=9200000 >/dev/null 2>/dev/null
					sysctl -w dev.nss.n2hcfg.n2h_high_water_core0=68904 >/dev/null 2>/dev/null
					sysctl -w dev.nss.n2hcfg.n2h_wifi_pool_buf=32768 >/dev/null 2>/dev/null
				fi
			;;
			ap-ac01)
				#total pbuf size is 160 bytes,allocate memory for 14712 pbufs
				sysctl -w dev.nss.n2hcfg.extra_pbuf_core0=2400000 >/dev/null 2>/dev/null
				sysctl -w dev.nss.n2hcfg.n2h_high_water_core0=26432 >/dev/null 2>/dev/null
				sysctl -w dev.nss.n2hcfg.n2h_wifi_pool_buf=0 >/dev/null 2>/dev/null

			;;
			ap-ac01)
				#total pbuf size is 160 bytes,allocate memory for 18808 pbufs
				sysctl -w dev.nss.n2hcfg.extra_pbuf_core0=3100000 >/dev/null 2>/dev/null
				sysctl -w dev.nss.n2hcfg.n2h_high_water_core0=30528 >/dev/null 2>/dev/null
				sysctl -w dev.nss.n2hcfg.n2h_wifi_pool_buf=4096 >/dev/null 2>/dev/null
			;;
			ap-hk* | ap-oak*)
				hk_ol_num="$(cat /lib/wifi/wifi_nss_hk_olnum)"
				[ -d /sys/module/qca_ol ] || { \
					if [ $hk_ol_num -eq 3 ]; then
						#total pbuf size is 160 bytes,allocate memory for 93560 pbufs
						#NSS general payload(8000),Rx Buffer per radio(4k),Tx queue buffer per radio(1k), intial TX allocation per radio(4k)
						#Below table is Tx desc allocation based on number of clients connected
						#Radio     Range0   Range1     Range2       Range3
						#           (<=64) (<=128)     (<=256)      (>256)
						#5G-Hi        24k	24k		24k		32k
						#2G           16k	16k		16k		16k
						#5G-Low       24k	24k		24k		32k
						#Absolute high water=NSS payloads + Rx buf per radio + Tx queue per radio + TxDescRange3(5g-low/5g-hi/2g)
						#wifi pool buff = Min(Total tx desc at range 3, device_limit) - total intial tx allocation
						#extra pbuf core0 = (high_water_core0 - (NSS + OCM buffers)) * pbuf_size
						#	where NSS+OCM buffers = 11720 and pbuf_size = 160
						# The need for correction of n2h_high_water_core0 from 105280 to 72512 is needed since
						#this high number can cause OOM
						sysctl -w dev.nss.n2hcfg.extra_pbuf_core0=10000000 >/dev/null 2>/dev/null
						sysctl -w dev.nss.n2hcfg.n2h_high_water_core0=72512 >/dev/null 2>/dev/null
						#initially after init 4k buf for 5G and 4k for 2G will be allocated, later range will be configured
						sysctl -w dev.nss.n2hcfg.n2h_wifi_pool_buf=36864 >/dev/null 2>/dev/null
					else	
						#total pbuf size is 160 bytes,allocate memory for 55672 pbufs
						sysctl -w dev.nss.n2hcfg.extra_pbuf_core0=9000000 >/dev/null 2>/dev/null
						sysctl -w dev.nss.n2hcfg.n2h_high_water_core0=67392 >/dev/null 2>/dev/null
						#initially after init 4k buf for 5G and 4k for 2G will be allocated, later range will be configured
						sysctl -w dev.nss.n2hcfg.n2h_wifi_pool_buf=40960 >/dev/null 2>/dev/null
					fi
				}
			;;
			ap-cp*)
				[ -d /sys/module/qca_ol ] || { \
					#total pbuf size is 160 bytes,allocate memory for 18808 pbufs
					sysctl -w dev.nss.n2hcfg.extra_pbuf_core0=3100000 >/dev/null 2>/dev/null
					sysctl -w dev.nss.n2hcfg.n2h_high_water_core0=30528 >/dev/null 2>/dev/null
					sysctl -w dev.nss.n2hcfg.n2h_wifi_pool_buf=4096 >/dev/null 2>/dev/null
				}
			;;
			*)
				#total pbuf size is 160 bytes,allocate memory for 48456 pbufs
				sysctl -w dev.nss.n2hcfg.extra_pbuf_core0=7800000 >/dev/null 2>/dev/null
				sysctl -w dev.nss.n2hcfg.n2h_high_water_core0=60176 >/dev/null 2>/dev/null
				sysctl -w dev.nss.n2hcfg.n2h_wifi_pool_buf=35840 >/dev/null 2>/dev/null
			;;
			esac
		fi
	fi

        if [ -n "$enable_mesh_support" ]; then
            config_get enable_mesh_peer_cap_update qcawifi enable_mesh_peer_cap_update
            [ -n "$enable_mesh_peer_cap_update" ] && append umac_args "enable_mesh_peer_cap_update=$enable_mesh_peer_cap_update"
        fi

        config_get enable_pktlog_support qcawifi enable_pktlog_support $def_pktlog_support
        [ -n "$enable_pktlog_support" ] && append umac_args "enable_pktlog_support=$enable_pktlog_support"

        config_get g_unicast_deauth_on_stop qcawifi g_unicast_deauth_on_stop $g_unicast_deauth_on_stop
        [ -n "$g_unicast_deauth_on_stop" ] && append umac_args "g_unicast_deauth_on_stop=$g_unicast_deauth_on_stop"

	config_get nss_wifi_olcfg qcawifi nss_wifi_olcfg
	if [ -n "$nss_wifi_olcfg" ]; then
		update_ini_file nss_wifi_olcfg "$nss_wifi_olcfg"
		config_get nss_wifi_nxthop_cfg qcawifi nss_wifi_nxthop_cfg
		if [ -n "$nss_wifi_nxthop_cfg" ]; then
		    update_ini_file nss_wifi_nxthop_cfg "$nss_wifi_nxthop_cfg"
		fi
	elif [ -f /lib/wifi/wifi_nss_olcfg ]; then
		nss_wifi_olcfg="$(cat /lib/wifi/wifi_nss_olcfg)"

		if [ $nss_wifi_olcfg != 0 ]; then
			if [ -f /lib/wifi/wifi_nss_override ] && [ $(cat /lib/wifi/wifi_nss_override) = 1 ]; then
				echo "NSS offload disabled due to unsupported config" >&2
				update_ini_file nss_wifi_olcfg 0
			else
				update_ini_file nss_wifi_olcfg "$nss_wifi_olcfg"
				update_ini_file rx_hash 0
			fi
		else
			update_ini_file nss_wifi_olcfg 0
		fi
	fi

	if [ -n "$nss_wifi_olcfg" ] && [ "$nss_wifi_olcfg" != "0" ]; then
		sysctl dev.nss.n2hcfg.n2h_high_water_core0 >/dev/null 2>/dev/null
	        nss_wifi_olnum="$(cat /lib/wifi/wifi_nss_olnum)"
        case "$board_name" in
        ap-hk*)
		[ -d /sys/module/qca_ol ] || { \
			#total pbuf size is 160 bytes,allocate memory for 35288 pbufs
			sysctl -w dev.nss.n2hcfg.extra_pbuf_core0=6000000 >/dev/null 2>/dev/null
			sysctl -w dev.nss.n2hcfg.n2h_high_water_core0=48128 >/dev/null 2>/dev/null
			sysctl -w dev.nss.n2hcfg.n2h_wifi_pool_buf=24576 >/dev/null 2>/dev/null
		}
        ;;
        *)
        	#total pbuf size is 160 bytes,allocate memory for 47680 pbufs
            sysctl -w dev.nss.n2hcfg.extra_pbuf_core0=8000000 >/dev/null 2>/dev/null
	    sysctl -w dev.nss.n2hcfg.n2h_high_water_core0=59392 >/dev/null 2>/dev/null
	    sysctl -w dev.nss.n2hcfg.n2h_wifi_pool_buf=35584 >/dev/null 2>/dev/null
        ;;
        esac
	fi

	wifiModParamFile="/tmp/wifi_mod_param.txt"
	[ -e "$wifiModParamFile" ] && rm "$wifiModParamFile" -rf

        obey=$(/usr/sbin/foreach wireless wifi-device obeyregpower 1)
        if [ -z "$obey" ]; then
                obey=$(/usr/sbin/foreach wireless wifi-device obeyregpower undefined)
                if [ -z "$obey" ]; then
                        obey=0
                else
                        obey=1
                fi
        else
                obey=1
        fi
	support_high_power=$(get_sn_wifi_option SUPPORT_HIGH_POWER)
	support_high_power_24g=$(get_sn_wifi_option SUPPORT_HIGH_POWER_24G)
	support_outdoor=$(get_sn_wifi_option SUPPORT_OUTDOOR_SLOTTIME_ACK_CTS_TIMEOUT)
	support_proc_in_umac=$(get_sn_wifi_option SUPPORT_PROC_IN_UMAC)

        [ -n "$obey" ] && [ "$support_high_power" == "1" ] && {
                append ol_args "ac_obey_reg_power=$obey"
                param=$(get_module_parameter "qca_ol" "ac_obey_reg_power")
                [ $param != "$obey" ] && reload=1
        }
        [ -n "$obey" ] && [ "$support_high_power_24g" == "1" ] && {
                append hal_args "ah_obey_reg_power=$obey"
                param=$(get_module_parameter "ath_hal" "ah_obey_reg_power")
                [ $param != "$obey" ] && reload=1
        }

	outdoor=$(uci -q get sysProductInfo.model.outdoor)
	[ -n "$outdoor" ] && [ "$support_outdoor" == "1" ] && {
		append ol_args "outdoor_device_setting=$outdoor"
		param=$(get_module_parameter "qca_ol" "outdoor_device_setting")
		[ $param != "$outdoor" ] && reload=1
	}

	atfmode=$(uci -q get airtime_fairness.atf.enabled)
	[ -n "$atfmode" ] && {
		append umac_args "atf_mode=$atfmode"
		param=$(get_module_parameter "umac" "atf_mode")
                [ $param != "$atfmode" ] && reload=1
	}

	reloadWifiMod=$(get_sn_wifi_option RELOAD_WIFI_KMODS)
	[ "$reloadWifiMod" == "0" ] && [ $reload == "1" ] && lock -u /var/run/wifilock &&unload_qcawifi	&& lock /var/run/wifilock #jacky: need to unload wifi modules if those parameter changed

	for mod in $(cat /lib/wifi/qca-wifi-modules); do
		case ${mod} in
			umac) [ -d /sys/module/${mod} ] || { \

				insmod ${mod} ${umac_args} || { \
					lock -u /var/run/wifilock
					unload_qcawifi
					return 1
				}
			};;

			qdf) [ -d /sys/module/${mod} ] || { \
				insmod ${mod} ${qdf_args} || { \
					lock -u /var/run/wifilock
					unload_qcawifi
					return 1
				}
			};;

			qca_ol) [ -f /tmp/no_qca_ol ] || { \
					[ -d /sys/module/${mod} ] || { \
					do_cold_boot_calibration
					insmod ${mod} ${ol_args} || { \
						lock -u /var/run/wifilock
						unload_qcawifi
						return 1
					}
				}
			};;

			ath_dev) [ -f /tmp/no_qca_da -a "$support_proc_in_umac" == "1" ] ||{ \
				[ -d /sys/module/${mod} ] || { \
					insmod ${mod} ${ath_dev_args} || { \
						lock -u /var/run/wifilock
						unload_qcawifi
						return 1
					}
				}
			};;

			ath_hal) [ -d /sys/module/${mod} ] || { \
					insmod ${mod} ${hal_args} || { \
						lock -u /var/run/wifilock
						unload_qcawifi
						return 1
				}
                        };;

			qca_da) [ -f /tmp/no_qca_da -a "$support_proc_in_umac" == "1" ] || { \
				[ -d /sys/module/${mod} ] || { \
					insmod ${mod} || { \
						lock -u /var/run/wifilock
						unload_qcawifi
						return 1
					}
				}
			};;

			#jacky: if support_proc_in_umac, we can follow qsdk qcawifi.sh that have no need to insmod those da modules
			hst_tx99|ath_rate_atheros) [ -f /tmp/no_qca_da -a "$support_proc_in_umac" == "1" ] || { \
				[ -d /sys/module/${mod} ] || { \
                                        insmod ${mod} || { \
                                                lock -u /var/run/wifilock
                                                unload_qcawifi
                                                return 1
                                        }
                                }
                        };;
			ath_pktlog) [ -n "$enable_pktlog_support" -a "$enable_pktlog_support" == "0" ] || { \
				[ -d /sys/module/${mod} ] || { \
					insmod ${mod} || { \
						lock -u /var/run/wifilock
						unload_qcawifi
						return 1
					}
				}
			};;

			*) [ -d /sys/module/${mod} ] || { \
				insmod ${mod} || { \
					lock -u /var/run/wifilock
					unload_qcawifi
					return 1
				}
			};;

		esac

		echo ${mod} ${args} >> "$wifiModParamFile"
	done

       # Remove DA/OL modules, if no DA/OL chipset found
	for dev in $(ls -d /sys/class/net/wifi* 2>&-); do
		[[ -f $dev/is_offload ]] || {
			qca_da_needed=1
		}
		[[ -f $dev/is_offload ]] && {
			qca_ol_needed=1
		}
	done

	if [ $qca_ol_needed -eq 0 ]; then
		if [ ! -f /tmp/no_qca_ol ]; then
			echo "No offload chipsets found." >/dev/console
			rmmod qca_ol > /dev/null 2> /dev/null
			cat "1" > /tmp/no_qca_ol
		fi
	fi

	if [ $qca_da_needed -eq 0 ]; then
		if [ ! -f /tmp/no_qca_da ]; then
			echo "No Direct-Attach chipsets found." >/dev/console
			rmmod qca_da > /dev/null 2> /dev/null
			if [ $support_proc_in_umac == "1" ]; then
				rmmod ath_dev > /dev/null 2> /dev/null
				rmmod hst_tx99 > /dev/null 2> /dev/null
				rmmod ath_rate_atheros > /dev/null 2> /dev/null
				rmmod ath_hal > /dev/null 2> /dev/null
			fi
			echo "1" > /tmp/no_qca_da
		fi
	fi
	
    #(X)jacky: seems no need to execute below flow, it is used on ipq806x
    #(O)SEANO George, spf11_csu1 needs to execute below script to improve throughput
	if [ -f "/lib/update_smp_affinity.sh" ]; then
		. /lib/update_smp_affinity.sh
		config_foreach enable_smp_affinity_wifi wifi-device
	fi

	lock -u /var/run/wifilock

	[ -f /etc/modules.d/50-batman-adv* ] && {
		for mod in $(cat /etc/modules.d/50-batman-adv* | sed 's/-/_/'); do
			[ -d /sys/module/${mod} ] || insmod $(mod)
		done
	}
}

unload_qcawifi() {
	#jacky: below codes are from spf10.0, but it's not suit for senao fw
	#config_load wireless
	#config_foreach disable_qcawifi wifi-device
	eval "type lowi_teardown" >/dev/null 2>&1 && lowi_teardown
	sleep 3
	killall cfr_test_app
	lock /var/run/wifilock
	for mod in $(cat /lib/wifi/qca-wifi-modules | sed '1!G;h;$!d'); do
		case ${mod} in
			mem_manager) continue;
		esac
		[ -d /sys/module/${mod} ] && rmmod ${mod}
	done
	lock -u /var/run/wifilock

	[ -e "/etc/modules.d/50-batman-adv*" ] && {
		for mod in $(cat /etc/modules.d/50-batman-adv* | sed 's/-/_/'); do      
			[ -d /sys/module/${mod} ] || rmmod ${mod}
		done
	}
}

destroy_vap() {
	local ifname="$1"
        ifconfig $ifname down
        wlanconfig $ifname destroy
}

reset_nawds_mode() {
	device=$1
	vif_strs="wds wds_0 mesh"	#jacky: temp using, maybe use wireless.$vif.nawds=1 later
	mesh_strs="ath33 ath34 ath36 ath38" #2.4g 5g temp_mesh 2.4g 5g scanning

	for str in $vif_strs;
	do
		vif=$device"_"$str
		ifname=$(uci get wireless.$vif.ifname)
		remove_mesh_if "$ifname"
		wlanconfig $ifname nawds mode 0
	done
	for m in $mesh_strs; do
		if_mac="$(ifconfig ${m} 2>/dev/null | grep HWaddr | awk -F " " '{printf $5}')"
		[ -n "$if_mac" ] && wlanconfig $m nawds mode 0
	done
}

disable_qcawifi() {
	local device="$1"
	local parent

	disable_fastscan
	disable_scan_radio "$device"
	disable_background_scan "$device"
	reset_nawds_mode "$device"	#spf10.0_PRE_CS mesh kernel panic issue, got to reset nawds mode before wifi down

	find_qcawifi_phy "$device" >/dev/null || return 1
	
	# disable_qcawifi also gets called for disabled radio during wifi up. Don't
	# remove the files if it gets from disabled radio.
	config_get disabled "$device" disabled
	if [ -f /tmp/wifi_nss_up_one_radio ] && [ "$disabled" = "0" ]; then
		rm /tmp/wifi_nss_up_one_radio
	fi

## pre-CAC
	# save nol file to working radio
	get_dfs_nol_list "$device" "$opmode"
## pre-CAC

	config_get phy "$device" phy

	set_wifi_down "$device"

	# in MBSS case, send notification to allow deletion of transmitting VAP
    # 2019.12.27 Senao George, it will cause kernel-panic when reloaing wifi. 
    # Workaround to fix.
    # 2019.12.30 Senao George, using QSDK FW. It can reach 5680 round wifi reloading
    # test. Maybe senao patch cause it fail(?)
	# device_if $device wifi_down_ind 1

	include /lib/network
	cd /sys/class/net
	for dev in *; do
		[ -f /sys/class/net/${dev}/parent ] && { \
			local parent=$(cat /sys/class/net/${dev}/parent)
			[ -n "$parent" -a "$parent" = "$device" ] && { \
				#new flow, to kill wifi related processes which use wpa_cli and global
				[ -f "/var/run/hostapd-${dev}.lock" ] && { \
					wpa_cli -g /var/run/hostapd/global raw REMOVE ${dev}
					rm /var/run/hostapd-${dev}.lock
				}
				[ -f "/var/run/wpa_supplicant-${dev}.lock" ] && { \
					wpa_cli -g /var/run/wpa_supplicantglobal  interface_remove  ${dev}
					rm /var/run/wpa_supplicant-${dev}.lock
				}
				[ -f "/var/run/wapid-${dev}.conf" ] && { \
					kill "$(cat "/var/run/wifi-${dev}.pid")"
				}
				#end

				#traditional flow to kill wifi related processes
				[ -f "/var/run/wifi-${dev}.pid" ] && { \
					kill "$(cat "/var/run/wifi-${dev}.pid")"
				}
				#end

				#unbridge "$dev" #jacky: it's empty in config.sh, do unbridge in notify_ubus_to_unbridge before interface down
				notify_ubus_to_unbridge "$dev"
				ifconfig "$dev" down
				if [ "$default_cfg80211"=="1" ]; then
					iw "$dev" del
				else
					wlanconfig "$dev" destroy
				fi
			}
			
			[ -f /var/run/hostapd_cred_${device}.bin ] && { \
				rm /var/run/hostapd_cred_${device}.bin
			}
		}
	done

	nrvaps=$(cat /sys/class/net/*/parent | grep "wifi" |wc -l)
	[ ${nrvaps} -gt 0 ] || {

		obey=$(/usr/sbin/foreach wireless wifi-device obeyregpower 1)                                                                                                      
                if [ -z "$obey" ]; then                                                                                                                                            
                        obey=0                                                                                                                                                     
                else                                                                                                                                                               
                        obey=1                                                                                                                                                     
                fi            

		reloadWifiMod=$(get_sn_wifi_option RELOAD_WIFI_KMODS)
		[ "$reloadWifiMod" == '1' -o "$obey" == "0" ] && unload_qcawifi
		# for traditional flow, remove it.
		hostapd_global_restart                  #Jacky: for hostapd connection refused issue, close hostapd global when wifi down
	}

	return 0
}


#
# Read ethaddr at /tmp/
# wifi0_mac_file="/tmp/wifi0_mac.txt"
# wifi1_mac_file="/tmp/wifi1_mac.txt"
# wifi2_mac_file="/tmp/wifi2_mac.txt"
#
set_mac()
{
	local wifi_dev=$1
	macaddr_file="/tmp/"$wifi_dev"_mac.txt"

	if [ -r $macaddr_file ]; then
		local macaddr=$(cat $macaddr_file)
		device_if $wifi_dev setHwaddr $macaddr 2>/dev/null
	else
		echo "qcawifi.sh can not read $macaddr_file" >> $error_msg
	fi
}

set_txpower()
{
# TXPower settings only work if device is up already
# while atheros hardware theoretically is capable of per-vif (even per-packet) txpower
# adjustment it does not work with the current atheros hal/madwifi driver
	local device=$1
	local vif=$2
	config_get txpower $device txpower
	config_get tpscale $device tpscale
	config_get obey $device obeyregpower
	config_get phy $device phy

	if [ -z "$obey" ]; then
		# If project doesn't support green_mode,
		# just need to set txpower.
		config_get vif_txpower "$vif" txpower
		txpower="${txpower:-$vif_txpower}"
		[ -z "$txpower" ] || iwconfig "$ifname" txpower "${txpower%%.*}"
		echo "$device no green mode" >> /tmp/noGreenMode.txt
		return 0
	fi

	# Check Config come from ezmcloud
	if [ ! -f /etc/config/ezmcloud ]; then
		# Not come from ezmcloud will run orignal method
		if [ "$txpower" == "0" ]; then
			# We always set maximum power 31(no value will be bigger than this value)
			# and driver will check ctl table to handle this value into correct power.
			[ "$obey" == "0" ] && txpower=31
		else
			[ "$obey" == "1" ] && txpower=0
		fi
	fi

	iwconfig "$ifname" txpower "${txpower%%.*}"

	if [ -n "$tpscale" ];then
		cfg80211tool "$device" tpscale "$tpscale"
	fi

}

cpu_irq_setup(){
	local wifi0_irq=$(cat /proc/interrupts | grep wifi0 | awk 'gsub(":","",$1) {print $1}')
	local wifi1_irq=$(cat /proc/interrupts | grep wifi1 | awk 'gsub(":","",$1) {print $1}')
	local wifi2_irq=$(cat /proc/interrupts | grep wifi2 | awk 'gsub(":","",$1) {print $1}')
	if [ -n "$wifi0_irq" ]; then
		if [ -f /proc/irq/${wifi0_irq}/smp_affinity ]; then
			echo 2 > /proc/irq/${wifi0_irq}/smp_affinity
		fi
	fi
	if [ -n "$wifi1_irq" ]; then
		if [ -f /proc/irq/${wifi1_irq}/smp_affinity ]; then
			echo 4 > /proc/irq/${wifi1_irq}/smp_affinity
		fi
	fi
	if [ -n "$wifi2_irq" ]; then
		if [ -f /proc/irq/${wifi2_irq}/smp_affinity ]; then
			echo 8 > /proc/irq/${wifi2_irq}/smp_affinity
		fi
	fi
}

enable_qcawifi() {
	local device="$1"
	[ -e "$ipq40xx_sh_file" ] && check_boarddata "$device"
	[ -f /tmp/sysinfo/board_name ] && {
	ax_board_name=$(cat /tmp/sysinfo/board_name)
	case "$ax_board_name" in
	ap-hk*|ap-ac01*|ap-cp*)
		#only for ax platform, ap-ac01* or ap-hk* or ap-cp* (cypress)
		prepare_ax_boarddata
		;;
	*)
		# not support
		;;
		esac
	}

	load_qcawifi
	cpu_irq_setup
	local setchannel=0;
	local domain=$(setconfig -g 4)
	local ac_mode=0;
	local num_radio_instamode=0
	local hk_ol_num=0
	local hwcaps
	local hwid=$(setconfig -g 9)

	if [ ! -f /lib/wifi/wifi_nss_override ]; then
		if [ -f /lib/wifi/wifi_nss_olcfg ] && [ $(cat /lib/wifi/wifi_nss_olcfg) != 0 ]; then
			touch /lib/wifi/wifi_nss_override
			echo 0 > /lib/wifi/wifi_nss_override
		fi
	fi

	if [ -f /lib/wifi/wifi_nss_override ]; then
		cd /sys/class/net
		for all_device in $(ls -d wifi* 2>&-); do
			config_get_bool disabled "$all_device" disabled 0
			[ $disabled = 0 ] || continue
			config_get vifs "$all_device" vifs

			for vif in $vifs; do
				config_get mode "$vif" mode
				if [ $mode = "sta" ]; then
					num_radio_instamode=$(($num_radio_instamode + 1))
					break
				fi
			done
			if [ $num_radio_instamode = "0" ]; then
				break
			fi
		done

		nss_override="$(cat /lib/wifi/wifi_nss_override)"
		if [ $num_radio_instamode = "3" ]; then
			config_get nss_wifi_olcfg qcawifi nss_wifi_olcfg
			if [ -n "$nss_wifi_olcfg" ] && [ $nss_wifi_olcfg != 0 ]; then
				echo " Invalid Configuration: 3 stations in offload not supported"
				return 1
			fi
			if [ $nss_override = "0" ]; then
				echo 1 > /lib/wifi/wifi_nss_override
				unload_qcawifi
				#device=$1	#jacky: not sure why need to assign device again, it makes wifi1 cannot be brought up
				load_qcawifi
			fi
		else
			if [ $nss_override != "0" ]; then
				echo 0 > /lib/wifi/wifi_nss_override
				unload_qcawifi
				#device=$1
				load_qcawifi
			fi
		fi
	fi

	enable_debug_wlan_script "$device"

	dfs_spoof "$device"

	green_mode "$device"

	set_mac "$device"

    # SENAO support generate vap mac before interface up
    [ -e "/usr/sbin/bsGetVapMac.sh" ] && bsGetVapMac.sh

	find_qcawifi_phy "$device" || return 0
	config_get phy "$device" phy

	# support SENAO outdoor distance parameter
    local outdoor=$(uci get sysProductInfo.model.outdoor)
    config_get distance "$device" distance
	[ -n "$distance" -a "$outdoor" = "1" ] && distance_func "$device" "$phy"

	#SENAO set default country, wireless.wifix.country may be 00 as default, need set it here by regdomain
	set_default_country "$device"

	config_get channel "$device" channel
	config_get vifs "$device" vifs

	[ "auto" == "$channel" ] && channel=0

	#set wifi params by radio, ex: device_if wifi1 xxxxx
	set_wifi_param_per_radio "$device"

	config_get opmode "$device" opmode
	
	case "$opmode" in
		ap|wds_ap|wds_bridge|sta_ap)
	
		#SENAO support channel config setting
		set_channel_config "$device"

		#SENAO support client limit setting
		set_client_limit "$device"	

		#SENAO support fasthandover
		set_sn_fasthandover "$device"

		config_get nochannel "$device" nochannel 0 # Is radio no channel supported

		[ "$nochannel" != "1" ] && {
			create_wifi_vaps "$device"

			#set wifi params by vap, ex: device_if athx xxxx
			set_wifi_param_per_vap "$device"

			#SENAO fix rate related issues
			if [ "$default_cfg80211" == "0" ]; then
				sn_fix_rate_issue_func "$device"	#not to use it on cfg80211
			fi

			#SENAO set channel after all vap set
			sn_set_channel_final "$device"

			#SENAO fix HT20_40 issue
			sn_fix_HT20_40_issue_func "$device"

			# fix QCA's bug, ap kernel panic issue.
			[ "$hwid" = "010100B5" -o "$hwid" = "010100B6" ] && fix_11ax_ap_panic_issue

			#SENAO enable fastscan feature
			#enable_fastscan "$device"	#old fastscan, should not be used anymore

			#SENAO set airtime fairness
			set_airtime_fairness_vip

			#SENAO set legacy deny
			set_legacy_deny "$device"

			set_beacon_interval "$device"	#set beacon interval makes connected sta since sending bcast deauth

			#set nawds mode or mesh
			set_nawds_or_mesh_mode "$device"

			#SENAO set vlan passthrough
			set_vlan_passthrough

			set_background_scan "$device"

			set_scan_radio "$device"
		}

		;;

		sta|wds_sta)
			set_channel_config "$device"
			create_wifi_vaps "$device"
			set_wifi_param_per_vap "$device"
		;;
		mon)
			#SENAO set scan radio
			create_wifi_vaps "$device"
                        set_scan_radio "$device"
		;;
	esac

	common_dev_setting_after_vap_up	"$device"
	#SENAO if no vap on the radio, create dummy vap
        create_dummy_interfaces "$device"

	# set_wifi_led "$device"
	echo 1 #without this makes _wifi_updown in /sbin/wifi shows "enable failed"

	#senao log get bridge mac address
	set_bridge_mac_to_proc

	set_rfdetector "$device"
}

pre_qcawifi() {
	local action=${1}

	config_load wireless

	config_get_bool sniffer_enable sniffer sniffer_enable 0
	echo ">>>>>sniffer_enable:$sniffer_enable" > /dev/console

	case "${action}" in
		disable)
			config_get_bool wps_vap_tie_dbdc qcawifi wps_vap_tie_dbdc 0

			if [ $wps_vap_tie_dbdc -ne 0 ]; then
				kill "$(cat "/var/run/hostapd.pid")"
				[ -f "/tmp/hostapd_conf_filename" ] &&
					rm /tmp/hostapd_conf_filename
			fi

			eval "type qwrap_teardown" >/dev/null 2>&1 && qwrap_teardown
			eval "type icm_teardown" >/dev/null 2>&1 && icm_teardown
			eval "type wpc_teardown" >/dev/null 2>&1 && wpc_teardown
			eval "type lowi_teardown" >/dev/null 2>&1 && lowi_teardown
			[ ! -f /etc/init.d/lbd ] || /etc/init.d/lbd stop
			[ ! -f /etc/init.d/hyd ] || /etc/init.d/hyd stop
			[ ! -f /etc/init.d/ssid_steering ] || /etc/init.d/ssid_steering stop
			[ ! -f /etc/init.d/mcsd ] || /etc/init.d/mcsd stop
			[ ! -f /etc/init.d/wsplcd ] || /etc/init.d/wsplcd stop

			rm -f /var/run/wifi-wps-enhc-extn.conf
			[ -r /var/run/wifi-wps-enhc-extn.pid ] && kill "$(cat "/var/run/wifi-wps-enhc-extn.pid")"
		;;
	esac

	wifiProfile2Config

	[ "$(uci -q get mesh.wifi.supportRP)" == "1" ] && mesh_RP_config
}

post_qcawifi() {
	local action=${1}
	config_get type "$device" type
	[ "$type" != "qcawifi" ] && return

	lock /var/run/wifilock
	case "${action}" in
		enable)
			local icm_enable qwrap_enable lowi_enable

			# Run a single hostapd instance for all the radio's
			# Enables WPS VAP TIE feature

			config_get_bool wps_vap_tie_dbdc qcawifi wps_vap_tie_dbdc 0

			if [ $wps_vap_tie_dbdc -ne 0 ]; then
				hostapd_conf_file=$(cat "/tmp/hostapd_conf_filename")
				hostapd -P /var/run/hostapd.pid $hostapd_conf_file -B
				config_foreach qcawifi_start_hostapd_cli wifi-device
			fi

			config_get_bool icm_enable icm enable 0
			[ ${icm_enable} -gt 0 ] && \
					eval "type icm_setup" >/dev/null 2>&1 && {
				icm_setup
			}

			config_get_bool wpc_enable wpc enable 0
			[ ${wpc_enable} -gt 0 ] && \
					eval "type wpc_setup" >/dev/null 2>&1 && {
				wpc_setup
			}

			config_get_bool lowi_enable lowi enable 0
			[ ${lowi_enable} -gt 0 ] && \
				eval "type lowi_setup" >/dev/null 2>&1 && {
				lowi_setup
			}

			eval "type qwrap_setup" >/dev/null 2>&1 && qwrap_setup && disable_qcawifi

			# These init scripts are assumed to check whether the feature is
			# actually enabled and do nothing if it is not.
			[ ! -f /etc/init.d/lbd ] || /etc/init.d/lbd start
			[ ! -f /etc/init.d/ssid_steering ] || /etc/init.d/ssid_steering start
			[ ! -f /etc/init.d/wsplcd ] || /etc/init.d/wsplcd restart

			config_get_bool wps_pbc_extender_enhance qcawifi wps_pbc_extender_enhance 0
			[ ${wps_pbc_extender_enhance} -ne 0 ] && { \
				rm -f /var/run/wifi-wps-enhc-extn.conf
				setup_wps_enhc
			}
			if [ -f  "/var/run/son.conf" ]; then
			    rm /var/run/son.conf
			fi

			config_load wireless
			config_foreach son_get_config wifi-device

			rm -f /etc/ath/iface_mgr.conf
			rm -f /var/run/iface_mgr.pid
			if [ "$(uci -q get mesh.wifi.supportRP)" != "1" ]; then
				iface_mgr_setup
			fi
			update_daemon_coldboot_qdss_support_variables
			qdss_post_setting
		;;
		enable_recover)
			[ ! -f /etc/init.d/lbd ] || /etc/init.d/lbd start
		;;
	esac

	atf_config
	set_wmm_dscp
	fix_fastroam
	set_throughput_rule

	lock -u /var/run/wifilock
}

check_qcawifi_device() {
	[ ${1%[0-9]} = "wifi" ] && config_set "$1" phy "$1"
	config_get phy "$1" phy
	[ -z "$phy" ] && {
		find_qcawifi_phy "$1" >/dev/null || return 0
		config_get phy "$1" phy
	}
	[ "$phy" = "$dev" ] && found=1
}

detect_qcawifi() {


	local enable_cfg80211=`uci show qcacfg80211.config.enable |grep "qcacfg80211.config.enable='1'"`
	[ -n "$enable_cfg80211" ] && echo "qcawifi configuration is disable" > /dev/console && return 1;

	is_ftm=`grep wifi_ftm_mode /proc/cmdline | wc -l`
	[ $is_ftm = 1 ] && ftm_qcawifi && return

	is_wal=`grep waltest_mode /proc/cmdline | wc -l`
	[ $is_wal = 1 ] && waltest_qcawifi && return

	if [ -e /sys/firmware/devicetree/base/MP_256 ]; then
		update_ini_for_lowmem QCA8074_i.ini
		update_ini_for_lowmem QCA8074V2_i.ini
		update_ini_for_lowmem QCA6018_i.ini
	fi

	config_present=0
	devidx=0
	socidx=0
	olcfg_ng=0
	olcfg_ac=0
	olcfg_axg=0
	olcfg_axa=0
	nss_olcfg=0
	nss_ol_num=0
	reload=0
	avoid_load=0
	hw_mode_detect=0
	num_chains=0
	is_e_build=0
	prefer_hw_mode_id="$(grep hw_mode_id \
			/ini/internal/global_i.ini | awk -F '=' '{print $2}')"

	[ $prefer_hw_mode_id -gt 6 ] && prefer_hw_mode_id=6

	if [ $prefer_hw_mode_id == 6 ]; then
		hw_mode_detect=1
	fi
	sleep 3

	if [ -e /sys/firmware/devicetree/base/MP_256 ]; then
		for mod in $(cat /lib/wifi/qca-wifi-modules); do
			case ${mod} in
				umac) [ -d /sys/module/${mod} ] && { \
					avoid_load=1
				};;
			esac
		done
	fi

	load_qcawifi
	config_load wireless
	local board_name

	[ -f /tmp/sysinfo/board_name ] && {
		board_name=$(cat /tmp/sysinfo/board_name)
	}

	#2020.06.08 Senao George, these code seems useless because config_present will stop
	#the use of $devidx
	#
	#while :; do
	#	config_get type "wifi$devidx" type
	#	[ -n "$type" ] || break
	#	devidx=$(($devidx + 1))
	#done
	cd /sys/class/net
	for soc in $(ls -d soc* 2>&-); do
		if [ -f ${soc}/hw_modes ]; then
			hw_modes=$(cat ${soc}/hw_modes)
			case "${hw_modes}" in
				*DBS_SBS:*)
					prefer_hw_mode_id=4;;
				*DBS:*)
					prefer_hw_mode_id=1;;
				*DBS_OR_SBS:*)
					prefer_hw_mode_id=5;;
				*SINGLE:*)
					prefer_hw_mode_id=0;;
				*SBS_PASSIVE:*)
					prefer_hw_mode_id=2;;
				*SBS:*)
					prefer_hw_mode_id=3;;
			esac
		fi
	done
	if [ $hw_mode_detect == 1 ]; then
		update_internal_ini global_i.ini hw_mode_id "$prefer_hw_mode_id"
	fi

	[ -d wifi0 ] || [ -d wifi1 ] || return
	for dev in $(ls -d wifi* 2>&-); do
		found=0
		#2020.06.08 Senao George, move to the end of the loop for processing the 
		#nss variabel when config existing.
		#
		#config_foreach check_qcawifi_device wifi-device
		#if [ "$found" -gt 0 ]; then
		#	config_present=1
		#       continue
		#fi

		if [ $num_chains -le $(cat ${dev}/txchains) ]; then
			num_chains=$(cat ${dev}/txchains)
		fi

		if [ $num_chains -le $(cat ${dev}/rxchains) ]; then
			num_chains=$(cat ${dev}/rxchains)
		fi

		[ -f ${dev}/is_e_build ] && {
			is_e_build=$(cat ${dev}/is_e_build)
		}

		hwcaps=$(cat ${dev}/hwcaps)
		case "${hwcaps}" in
			*11bgn) mode_11=ng;;
			*11bgn/ax) mode_11=axg;;
			*11abgn) mode_11=ng;;
			*11an) mode_11=na;;
			*11an/ac) mode_11=ac;;
			*11an/ac/ax) mode_11=axa;;
			*11abgn/ac) mode_11=ac;;
			*11abgn/ac/ax) mode_11=axa;;
		esac
		if [ -f /sys/class/net/${dev}/nssoffload ] && [ $(cat /sys/class/net/${dev}/nssoffload) == "capable" ]; then
			case "${mode_11}" in
				ng)
					if [ $olcfg_ng == 0 ]; then
						olcfg_ng=1
						nss_olcfg=$(($nss_olcfg|$((1<<$devidx))))
						nss_ol_num=$(($nss_ol_num + 1))
					fi
				;;
				na|ac)
					if [ $olcfg_ac == 0 ]; then
						olcfg_ac=1
						nss_olcfg=$(($nss_olcfg|$((1<<$devidx))))
						nss_ol_num=$(($nss_ol_num + 1))
					fi
				;;
				axa)
					if [ $olcfg_axa -le 1 ]; then
						olcfg_axa=$(($olcfg_axa + 1))
						nss_olcfg=$(($nss_olcfg|$((1<<$devidx))))
						nss_ol_num=$(($nss_ol_num + 1))
					fi
				;;
				axg)
					if [ $olcfg_axg == 0 ]; then
						olcfg_axg=1
						nss_olcfg=$(($nss_olcfg|$((1<<$devidx))))
						nss_ol_num=$(($nss_ol_num + 1))
					fi
				;;
			esac
		reload=1
		fi

		config_foreach check_qcawifi_device wifi-device
		if [ "$found" -gt 0 ]; then
			echo ">>>>>wifi$devidx exist:[nss_olcfg:$nss_olcfg nss_ol_num:$nss_ol_num]" > /dev/console
			config_present=0
			devidx=$(($devidx + 1))
		       	continue
		fi

		cat <<EOF
config wifi-device  wifi$devidx
	option type	qcawifi
	option channel	auto
	option macaddr	$(cat /sys/class/net/${dev}/address)
	option hwmode	11${mode_11}
	option opmode	ap
	# REMOVE THIS LINE TO ENABLE WIFI:
	option disabled 1

config wifi-iface
	option device	wifi$devidx
	option network	lan
	option mode	ap
	option mode_display	ap
	option ssid	OpenWrt
	option ifname	ath$devidx
	option encryption none

EOF
	devidx=$(($devidx + 1))
	done

	#config_present 1 indicates that /etc/config/wireless is already having some configuration.
	# In that case we shall not update the olcfg files
	if [ $config_present == 0 ]; then
		case "$board_name" in
		ap-dk01.1-c1 | ap-dk01.1-c2 | ap-dk04.1-c1 | ap-dk04.1-c2 | ap-dk04.1-c3 | ap152 | ap147 | ap151 | ap135 | ap137)
			;;
		ap-hk*|ap-ac*|ap-oa*)
			if [ -f /etc/rc.d/*qca-nss-ecm ]; then
				echo $nss_olcfg >/lib/wifi/wifi_nss_olcfg
				echo $nss_ol_num >/lib/wifi/wifi_nss_olnum
				echo "$(($olcfg_axa + $olcfg_axg))" > /lib/wifi/wifi_nss_hk_olnum
			else
				#echo 0 >/lib/wifi/wifi_nss_olcfg
				#2020.12.7 Senao George, fix mpsk vlan cannot be greater than 255.
				echo $nss_olcfg >/lib/wifi/wifi_nss_olcfg
				echo $nss_ol_num >/lib/wifi/wifi_nss_olnum
				echo "$(($olcfg_axa + $olcfg_axg))" > /lib/wifi/wifi_nss_hk_olnum
			fi
			;;
		*)
			echo $nss_olcfg >/lib/wifi/wifi_nss_olcfg
			echo $nss_ol_num >/lib/wifi/wifi_nss_olnum
			;;
		esac

		if [ ! -f /sys/firmware/devicetree/base/MP_256 ]; then
			if [ $num_chains -le 4 ]; then
				update_ini_for_monitor_buf_ring QCA8074_i.ini
				update_ini_for_monitor_buf_ring QCA8074V2_i.ini
				update_ini_for_monitor_buf_ring QCA6018_i.ini
			fi
		fi

		if [ -e /sys/firmware/devicetree/base/MP_512 ]; then
			update_ini_for_512MP_dp_tx_desc QCA8074_i.ini
			update_ini_for_512MP_dp_tx_desc QCA8074V2_i.ini
			update_ini_for_512MP_dp_tx_desc QCA6018_i.ini
			if [ $is_e_build -ne 1 ]; then
				update_ini_for_512MP_P_build QCA8074_i.ini
				update_ini_for_512MP_P_build QCA8074V2_i.ini
				update_ini_for_512MP_P_build QCA6018_i.ini
			fi
		fi
	fi
	sync

	if [ $reload == 1 ] ; then
		if [ $avoid_load == 1 ]; then
			wifi_updown "disable" "$2" > /dev/null
			ubus call network reload
			wifi_updown "enable" "$2" > /dev/null
		else
			unload_qcawifi > /dev/null
			load_qcawifi > /dev/null
		fi
	fi

	start_recovery_daemon
}

enable_debug_wlan_script() {
        #debug wlan-script mode
	local wifi_dev=$1
        local dbglvl=$(setconfig -g 10)
        if [ "$dbglvl" -eq 2 ]; then
                case "$wifi_dev" in
                        wifi0)
                                dl.sh -d 2
                                sh /tmp/athwlan0.sh
                        ;;
                        wifi1)
                                dl.sh -d 3
                                sh /tmp/athwlan1.sh
                        ;;
                        wifi2)
                                dl.sh -d 3
                                sh /tmp/athwlan2.sh
                        ;;
                esac
                return;
        fi
}

set_wifi_param_per_radio() {

	local wifi_dev=$1
	config_get phy "$wifi_dev" phy

    ###2020.06.05 Senao George, Add missing config from original qsdk script
	config_get bsta_fixed_idmask "$device" bsta_fixed_idmask 255
	[ -n "$bsta_fixed_idmask" ] && device_if "$phy" bsta_fixed_idmask "$bsta_fixed_idmask"

    ###2020.10.30 Senao George, porting from qca_spf_11_0_csu3
    config_get chan_coex_disable "$device" chan_coex_disable
    [ -n "$chan_coex_disable" ] && device_if "$phy" chan_coex_disable "$chan_coex_disable"

    config_get ext_nss "$device" ext_nss
    case "$ext_nss" in
        1|on|enabled) device_if "$phy" ext_nss 1 >/dev/null 2>&1
            ;;
        0|on|enabled) device_if "$phy" ext_nss 0 >/dev/null 2>&1
            ;;
        *) ;;
    esac

    config_get_bool bcnburst "$device" bcnburst 1
    [ -n "$bcnburst" ] && device_if "$phy" set_bcnburst "$bcnburst"

	config_get obss_rxrssi_th "$device" obss_rxrssi_th 35
	[ -n "$obss_rxrssi_th" ] && device_if "$phy" obss_rxrssi_th "${obss_rxrssi_th}"

        config_get interCACChan "$device" interCACChan
        [ -n "$interCACChan" ] && device_if "$phy" interCACChan "$interCACChan"

        config_get preCACEn "$device" preCACEn
        [ -n "$preCACEn" ] && device_if "$phy" preCACEn "$preCACEn"

	config_get pCACTimeout "$device" pCACTimeout
	[ -n "$pCACTimeout" ] && device_if "$phy" pCACTimeout "$pCACTimeout"

        config_get regdomain "$device" regdomain
        [ -n "$regdomain" ] && device_if "$phy" setRegdomain "$regdomain"

        config_get AMPDU "$device" AMPDU
        [ -n "$AMPDU" ] && device_if "$phy" AMPDU "$AMPDU"

        config_get ampdudensity "$device" ampdudensity
        [ -n "$ampdudensity" ] && device_if "$phy" ampdudensity "$ampdudensity"

        config_get_bool AMSDU "$device" AMSDU
        [ -n "$AMSDU" ] && device_if "$phy" AMSDU "$AMSDU"

        config_get AMPDULim "$device" AMPDULim
        [ -n "$AMPDULim" ] && device_if "$phy" AMPDULim "$AMPDULim"

        config_get AMPDUFrames "$device" AMPDUFrames
        [ -n "$AMPDUFrames" ] && device_if "$phy" AMPDUFrames "$AMPDUFrames"

        config_get AMPDURxBsize "$device" AMPDURxBsize
        [ -n "$AMPDURxBsize" ] && device_if "$phy" AMPDURxBsize "$AMPDURxBsize"

        config_get set_smart_antenna "$device" set_smart_antenna
        [ -n "$set_smart_antenna" ] && device_if "$phy" setSmartAntenna "$set_smart_antenna"

        config_get current_ant "$device" current_ant
        [ -n  "$current_ant" ] && device_if "$phy" current_ant "$current_ant"

        config_get default_ant "$device" default_ant
        [ -n "$default_ant" ] && device_if "$phy" default_ant "$default_ant"

        config_get ant_retrain "$device" ant_retrain
        [ -n "$ant_retrain" ] && device_if "$phy" ant_retrain "$ant_retrain"

        config_get retrain_interval "$device" retrain_interval
        [ -n "$retrain_interval" ] && device_if "$phy" retrain_interval "$retrain_interval"

        config_get retrain_drop "$device" retrain_drop
        [ -n "$retrain_drop" ] && device_if "$phy" retrain_drop "$retrain_drop"

        config_get ant_train "$device" ant_train
        [ -n "$ant_train" ] && device_if "$phy" ant_train "$ant_train"

        config_get ant_trainmode "$device" ant_trainmode
        [ -n "$ant_trainmode" ] && device_if "$phy" ant_trainmode "$ant_trainmode"

        config_get ant_traintype "$device" ant_traintype
        [ -n "$ant_traintype" ] && device_if "$phy" ant_traintype "$ant_traintype"

        config_get ant_pktlen "$device" ant_pktlen
        [ -n "$ant_pktlen" ] && device_if "$phy" ant_pktlen "$ant_pktlen"

        config_get ant_numpkts "$device" ant_numpkts
        [ -n "$ant_numpkts" ] && device_if "$phy" ant_numpkts "$ant_numpkts"

        config_get ant_numitr "$device" ant_numitr
        [ -n "$ant_numitr" ] && device_if "$phy" ant_numitr "$ant_numitr"

        config_get ant_train_thres "$device" ant_train_thres
        [ -n "$ant_train_thres" ] && device_if "$phy" train_threshold "$ant_train_thres"

        config_get ant_train_min_thres "$device" ant_train_min_thres
        [ -n "$ant_train_min_thres" ] && device_if "$phy" train_threshold "$ant_train_min_thres"

        config_get ant_traffic_timer "$device" ant_traffic_timer
        [ -n "$ant_traffic_timer" ] && device_if "$phy" traffic_timer "$ant_traffic_timer"

        config_get dcs_coch_int "$device" dcs_coch_int
        [ -n "$dcs_coch_int" ] && device_if "$phy" set_dcs_coch_int "$dcs_coch_int"

        config_get dcs_errth "$device" dcs_errth
        [ -n "$dcs_errth" ] && device_if "$phy" set_dcs_errth "$dcs_errth"

        config_get dcs_phyerrth "$device" dcs_phyerrth
        [ -n "$dcs_phyerrth" ] && device_if "$phy" set_dcs_phyerrth "$dcs_phyerrth"

        config_get dcs_usermaxc "$device" dcs_usermaxc
        [ -n "$dcs_usermaxc" ] && device_if "$phy" set_dcs_usermaxc "$dcs_usermaxc"

        config_get dcs_debug "$device" dcs_debug
        [ -n "$dcs_debug" ] && device_if "$phy" set_dcs_debug "$dcs_debug"

        config_get set_ch_144 "$device" set_ch_144
        [ -n "$set_ch_144" ] && device_if "$phy" setCH144 "$set_ch_144"

        config_get eppovrd_ch_144 "$device" eppovrd_ch_144
        [ -n "$eppovrd_ch_144" ] && device_if "$phy" setCH144EppOvrd "$eppovrd_ch_144"

        config_get_bool ani_enable "$device" ani_enable
        [ -n "$ani_enable" ] && device_if "$phy" ani_enable "$ani_enable"

        config_get_bool acs_bkscanen "$device" acs_bkscanen
        [ -n "$acs_bkscanen" ] && device_if "$phy" acs_bkscanen "$acs_bkscanen"

        config_get acs_scanintvl "$device" acs_scanintvl
        [ -n "$acs_scanintvl" ] && device_if "$phy" acs_scanintvl "$acs_scanintvl"

        config_get acs_rssivar "$device" acs_rssivar
        [ -n "$acs_rssivar" ] && device_if "$phy" acs_rssivar "$acs_rssivar"

        config_get acs_chloadvar "$device" acs_chloadvar
        [ -n "$acs_chloadvar" ] && device_if "$phy" acs_chloadvar "$acs_chloadvar"

        config_get acs_lmtobss "$device" acs_lmtobss
        [ -n "$acs_lmtobss" ] && device_if "$phy" acs_lmtobss "$acs_lmtobss"

        config_get acs_ctrlflags "$device" acs_ctrlflags
        [ -n "$acs_ctrlflags" ] && device_if "$phy" acs_ctrlflags "$acs_ctrlflags"

        config_get acs_dbgtrace "$device" acs_dbgtrace
        [ -n "$acs_dbgtrace" ] && device_if "$phy" acs_dbgtrace "$acs_dbgtrace"

        config_get_bool dscp_ovride "$device" dscp_ovride
        [ -n "$dscp_ovride" ] && device_if "$phy" set_dscp_ovride "$dscp_ovride"

        config_get reset_dscp_map "$device" reset_dscp_map
        [ -n "$reset_dscp_map" ] && device_if "$phy" reset_dscp_map "$reset_dscp_map"

        config_get dscp_tid_map "$device" dscp_tid_map
        [ -n "$dscp_tid_map" ] && device_if "$phy" set_dscp_tid_map $dscp_tid_map

        #Default enable IGMP overide & TID=6
        device_if "$phy" sIgmpDscpOvrid 1
        device_if "$phy" sIgmpDscpTidMap 6

        config_get_bool igmp_dscp_ovride "$device" igmp_dscp_ovride
        [ -n "$igmp_dscp_ovride" ] && device_if "$phy" sIgmpDscpOvrid "$igmp_dscp_ovride"

        config_get igmp_dscp_tid_map "$device" igmp_dscp_tid_map
        [ -n "$igmp_dscp_tid_map" ] && device_if "$phy" sIgmpDscpTidMap "$igmp_dscp_tid_map"

        config_get_bool hmmc_dscp_ovride "$device" hmmc_dscp_ovride
        [ -n "$hmmc_dscp_ovride" ] && device_if "$phy" sHmmcDscpOvrid "$hmmc_dscp_ovride"

        config_get hmmc_dscp_tid_map "$device" hmmc_dscp_tid_map
        [ -n "$hmmc_dscp_tid_map" ] && device_if "$phy" sHmmcDscpTidMap "$hmmc_dscp_tid_map"

        config_get_bool blk_report_fld "$device" blk_report_fld
        [ -n "$blk_report_fld" ] && device_if "$phy" setBlkReportFld "$blk_report_fld"

        config_get_bool drop_sta_query "$device" drop_sta_query
        [ -n "$drop_sta_query" ] && device_if "$phy" setDropSTAQuery "$drop_sta_query"

        config_get burst_dur "$device" burst_dur
        [ -n "$burst_dur" ] && device_if "$phy" burst_dur "$burst_dur"

        config_get TXPowLim2G "$device" TXPowLim2G
        [ -n "$TXPowLim2G" ] && device_if "$phy" TXPowLim2G "$TXPowLim2G"

        config_get TXPowLim5G "$device" TXPowLim5G
        [ -n "$TXPowLim5G" ] && device_if "$phy" TXPowLim5G "$TXPowLim5G"

        [ -f /tmp/sysinfo/board_name ] && {
                board_name=$(cat /tmp/sysinfo/board_name)
        }

        case "$board_name" in
	ap-hk*|ap-ac*|ap-cp*|ap-oak*)
		echo "Enable ol_stats by default for Lithium platforms"
		device_if "$phy" enable_ol_stats 1
        ;;
        *)
		echo "ol_stats is disabled by default for Non Lithium platforms"
        ;;
        esac

        config_get_bool enable_ol_stats "$device" enable_ol_stats
        [ -n "$enable_ol_stats" ] && device_if "$phy" enable_ol_stats "$enable_ol_stats"

        config_get emiwar80p80 "$device" emiwar80p80
        [ -n "$emiwar80p80" ] && device_if "$phy" emiwar80p80 "$emiwar80p80"

        config_get_bool rst_tso_stats "$device" rst_tso_stats
        [ -n "$rst_tso_stats" ] && device_if "$phy" rst_tso_stats "$rst_tso_stats"

        config_get_bool rst_lro_stats "$device" rst_lro_stats
        [ -n "$rst_lro_stats" ] && device_if "$phy" rst_lro_stats "$rst_lro_stats"

        config_get_bool rst_sg_stats "$device" rst_sg_stats
        [ -n "$rst_sg_stats" ] && device_if "$phy" rst_sg_stats "$rst_sg_stats"

        config_get_bool set_fw_recovery "$device" set_fw_recovery
        [ -n "$set_fw_recovery" ] && device_if "$phy" set_fw_recovery "$set_fw_recovery"

        config_get_bool allowpromisc "$device" allowpromisc
        [ -n "$allowpromisc" ] && device_if "$phy" allowpromisc "$allowpromisc"

        config_get set_sa_param "$device" set_sa_param
        [ -n "$set_sa_param" ] && device_if "$phy" set_sa_param $set_sa_param

        config_get_bool aldstats "$device" aldstats
        [ -n "$aldstats" ] && device_if "$phy" aldstats "$aldstats"

        config_get macaddr "$device" macaddr
        [ -n "$macaddr" ] && device_if "$phy" setHwaddr "$macaddr"

        config_get promisc "$device" promisc
        [ -n "$promisc" ] && device_if "$phy" promisc $promisc

        config_get mode0 "$device" mode0
        [ -n "$mode0" ] && device_if "$phy" fc_buf_min 2501

        config_get mode1 "$device" mode1
        [ -n "$mode1" ] && device_if "$phy" fc_buf_min 0

        handle_aggr_burst() {
                local value="$1"
                [ -n "$value" ] && device_if "$phy" aggr_burst $value
        }

        config_list_foreach "$device" aggr_burst handle_aggr_burst

        config_get_bool block_interbss "$device" block_interbss
        [ -n "$block_interbss" ] && device_if "$phy" block_interbss "$block_interbss"

        config_get set_pmf "$device" set_pmf
        [ -n "$set_pmf" ] && device_if "$phy" set_pmf "${set_pmf}"

        config_get txbf_snd_int "$device" txbf_snd_int 100
        [ -n "$txbf_snd_int" ] && device_if "$phy" txbf_snd_int "$txbf_snd_int"

        config_get mcast_echo "$device" mcast_echo
        [ -n "$mcast_echo" ] && device_if "$phy" mcast_echo "${mcast_echo}"

        config_get obss_rssi_th "$device" obss_rssi_th 35
        [ -n "$obss_rssi_th" ] && device_if "$phy" obss_rssi_th "${obss_rssi_th}"

        # Remove default value.
        # Since the obss_rxrssi_th value is too high,
        # the bandwidth of the DUT can not be reduced to HT20
        # when the STA sends a packet with the capability of "40 MHz Intolerantdut" or "20 MHz BSS Width Request"
        # in the 20/40 Mhz Coexistence test.
        # Please refer to the document: "Section 4.2.41 AP 20/40 MHz Coexistence of Wi-Fi CERTIFIED n Test Plan v2.0.11".
        config_get obss_rxrssi_th "$device" obss_rxrssi_th
        [ -n "$obss_rxrssi_th" ] && device_if "$phy" obss_rxrssi_th "${obss_rxrssi_th}"

        config_get acs_txpwr_opt "$device" acs_txpwr_opt
        [ -n "$acs_txpwr_opt" ] && device_if "$phy" acs_tcpwr_opt "${acs_txpwr_opt}"

        config_get obss_long_slot "$device" obss_long_slot
        [ -n "$obss_long_slot" ] && device_if "$phy" obss_long_slot "${obss_long_slot}"

        config_get staDFSEn "$device" staDFSEn
        [ -n "$staDFSEn" ] && device_if "$phy" staDFSEn "${staDFSEn}"

        config_get dbdc_enable "$device" dbdc_enable
        [ -n "$dbdc_enable" ] && device_if "$phy" dbdc_enable "${dbdc_enable}"

        config_get client_mcast "$device" client_mcast
        [ -n "$client_mcast" ] && device_if "$phy" client_mcast "${client_mcast}"

        config_get pas_scanen "$device" pas_scanen
        [ -n "$pas_scanen" ] && device_if "$phy" pas_scanen "${pas_scanen}"

        config_get delay_stavapup "$device" delay_stavapup
        [ -n "$delay_stavapup" ] && device_if "$phy" delay_stavapup "${delay_stavapup}"

        config_get tid_override_queue_map "$device" tid_override_queue_map
        [ -n "$tid_override_queue_map" ] && device_if "$phy" queue_map "${tid_override_queue_map}"

        config_get channel_block_mode "$device" channel_block_mode
        [ -n "$channel_block_mode" ] && device_if "$phy" acs_bmode "${channel_block_mode}"

        config_get no_vlan "$device" no_vlan
        [ -n "$no_vlan" ] && device_if "$phy" no_vlan "${no_vlan}"

        config_get discon_time qcawifi discon_time 10
        [ -n "$discon_time" ] && device_if "$phy" discon_time "${discon_time}"

        config_get reconfig_time qcawifi reconfig_time 60
        [ -n "$reconfig_time" ] && device_if "$phy" reconfig_time "${reconfig_time}"

        config_get alwaysprimary qcawifi alwaysprimary
        [ -n "$alwaysprimary" ] && device_if "$phy" alwaysprimary "${alwaysprimary}"

        config_get samessid_disable qcawifi samessid_disable
        [ -n "$samessid_disable" ] && device_if "$phy" samessid_disable "${samessid_disable}"

        config_get nss_wifi_olcfg qcawifi nss_wifi_olcfg
        if [ -z "$nss_wifi_olcfg" ]; then
                if [ -f /lib/wifi/wifi_nss_olcfg ]; then
                        nss_wifi_olcfg="$(cat /lib/wifi/wifi_nss_olcfg)"
                fi
        fi

	if [ -n "$nss_wifi_olcfg" ] && [ "$nss_wifi_olcfg" != "0" ]; then
		local mp_256="$(ls /proc/device-tree/ | grep -rw "MP_256")"
		local mp_512="$(ls /proc/device-tree/ | grep -rw "MP_512")"
		config_get hwmode "$device" hwmode auto
		hk_ol_num="$(cat /lib/wifi/wifi_nss_hk_olnum)"
		#For 256 memory profile the range is preset in fw
		if [ "$mp_256" == "MP_256" ]; then
			:
		elif [ "$mp_512" == "MP_512" ]; then
			if [ $hk_ol_num -eq 3 ]; then
				if [ ! -f /lib/wifi/wifi_nss_numradios ]; then
					touch /tmp/wifi_nss_up_one_radio
					sysctl -w dev.nss.n2hcfg.n2h_high_water_core0=31648 >/dev/null 2>/dev/null
					sysctl -w dev.nss.n2hcfg.n2h_wifi_pool_buf=0 >/dev/null 2>/dev/null
				fi
				case "$hwmode" in
				*axa | *ac)
					device_if "$phy" fc_buf0_max 4096
					device_if "$phy" fc_buf1_max 4096
					device_if "$phy" fc_buf2_max 4096
					device_if "$phy" fc_buf3_max 4096
					;;
				*)
					device_if "$phy" fc_buf0_max 4096
					device_if "$phy" fc_buf1_max 4096
					device_if "$phy" fc_buf2_max 4096
					device_if "$phy" fc_buf3_max 4096
					;;
				esac
			else
				if [ ! -f /tmp/wifi_nss_up_one_radio ]; then
					touch /tmp/wifi_nss_up_one_radio
					sysctl -w dev.nss.n2hcfg.n2h_high_water_core0=30624 >/dev/null 2>/dev/null
					sysctl -w dev.nss.n2hcfg.n2h_wifi_pool_buf=8192 >/dev/null 2>/dev/null
				fi
				case "$hwmode" in
				*axa | *ac)
					device_if "$phy" fc_buf0_max 8192
					device_if "$phy" fc_buf1_max 8192
					device_if "$phy" fc_buf2_max 8192
					device_if "$phy" fc_buf3_max 8192
					;;
				*)
					device_if "$phy" fc_buf0_max 8192
					device_if "$phy" fc_buf1_max 8192
					device_if "$phy" fc_buf2_max 8192
					device_if "$phy" fc_buf3_max 8192
					;;
				esac
			fi
		else
			case "$board_name" in
			ap-ac01)
				;;
			ap-ac02)
				case "$hwmode" in
				*axa | *ac)
					device_if "$phy" fc_buf0_max 4096
					device_if "$phy" fc_buf1_max 8192
					device_if "$phy" fc_buf2_max 8192
					device_if "$phy" fc_buf3_max 8192
					;;
				*)
					device_if "$phy" fc_buf0_max 4096
					device_if "$phy" fc_buf1_max 4096
					device_if "$phy" fc_buf2_max 4096
					device_if "$phy" fc_buf3_max 4096
					;;
				esac
				;;
			ap-cp*)
				if [ ! -f /tmp/wifi_nss_up_one_radio ]; then
					touch /tmp/wifi_nss_up_one_radio
					#reset the high water mark for NSS if range 0 value > 4096
					sysctl -w dev.nss.n2hcfg.n2h_high_water_core0=30528 >/dev/null 2>/dev/null
					#initially after init 4k buf for 5G and 4k for 2G will be allocated, later range will be configured
					sysctl -w dev.nss.n2hcfg.n2h_wifi_pool_buf=4096 >/dev/null 2>/dev/null
				fi
				case "$hwmode" in
				*axa | *ac)
					device_if "$phy" fc_buf0_max 8192
					device_if "$phy" fc_buf1_max 8192
					device_if "$phy" fc_buf2_max 8192
					device_if "$phy" fc_buf3_max 8192
				;;
				*)
					device_if "$phy" fc_buf0_max 4096
					device_if "$phy" fc_buf1_max 4096
					device_if "$phy" fc_buf2_max 4096
					device_if "$phy" fc_buf3_max 4096
				;;
				esac
			;;
			ap-hk09*)
				local soc_version_major
				[ -f /sys/firmware/devicetree/base/soc_version_major ] && {
					soc_version_major="$(hexdump -n 1 -e '"%1d"' /sys/firmware/devicetree/base/soc_version_major)"
				}

				if [ $soc_version_major = 2 ];then
					if [ ! -f /tmp/wifi_nss_up_one_radio ]; then
						touch /tmp/wifi_nss_up_one_radio
						#reset the high water mark for NSS if range 0 value > 4096
						sysctl -w dev.nss.n2hcfg.n2h_high_water_core0=67392 >/dev/null 2>/dev/null
						#initially after init 4k buf for 5G and 4k for 2G will be allocated, later range will be configured
						sysctl -w dev.nss.n2hcfg.n2h_wifi_pool_buf=40960 >/dev/null 2>/dev/null
					fi
					case "$hwmode" in
					*axa | *ac)
						device_if "$phy" fc_buf0_max 32768
						device_if "$phy" fc_buf1_max 32768
						device_if "$phy" fc_buf2_max 32768
						device_if "$phy" fc_buf3_max 32768
					;;
					*)
						device_if "$phy" fc_buf0_max 16384
						device_if "$phy" fc_buf1_max 16384
						device_if "$phy" fc_buf2_max 16384
						device_if "$phy" fc_buf3_max 32768
					;;
					esac
				else
					case "$hwmode" in
					*ac)
						#we distinguish the legacy chipset based on the hwcaps
						hwcaps=$(cat /sys/class/net/${phy}/hwcaps)
						if [ "$hwcaps" == "802.11an/ac" ]; then
							device_if "$phy" fc_buf0_max 8192
							device_if "$phy" fc_buf1_max 12288
							device_if "$phy" fc_buf2_max 16384
						else
							device_if "$phy" fc_buf0_max 4096
							device_if "$phy" fc_buf1_max 8192
							device_if "$phy" fc_buf2_max 12288
						fi
						device_if "$phy" fc_buf3_max 16384
						;;
					*)
						device_if "$phy" fc_buf0_max 4096
						device_if "$phy" fc_buf1_max 8192
						device_if "$phy" fc_buf2_max 12288
						device_if "$phy" fc_buf3_max 16384
					;;
					esac
				fi
			;;
			ap-hk* | ap-oak*)
				if [ $hk_ol_num -eq 3 ]; then
					if [ ! -f /tmp/wifi_nss_up_one_radio ]; then
						touch /tmp/wifi_nss_up_one_radio
						sysctl -w dev.nss.n2hcfg.n2h_high_water_core0=72512 >/dev/null 2>/dev/null
						sysctl -w dev.nss.n2hcfg.n2h_wifi_pool_buf=36864 >/dev/null 2>/dev/null
						update_ini_for_hk_sbs QCA8074V2_i.ini
					fi
					case "$hwmode" in
					*axa | *ac)
						device_if "$phy" fc_buf0_max 24576
						device_if "$phy" fc_buf1_max 24576
						device_if "$phy" fc_buf2_max 24576
						device_if "$phy" fc_buf3_max 32768
						;;
					*)
						device_if "$phy" fc_buf0_max 16384
						device_if "$phy" fc_buf1_max 16384
						device_if "$phy" fc_buf2_max 16384
						device_if "$phy" fc_buf3_max 24576
						;;
					esac
				else
					local soc_version_major
					[ -f /sys/firmware/devicetree/base/soc_version_major ] && {
						soc_version_major="$(hexdump -n 1 -e '"%1d"' /sys/firmware/devicetree/base/soc_version_major)"
					}

					if [ ! -f /tmp/wifi_nss_up_one_radio ]; then
						touch /tmp/wifi_nss_up_one_radio
						#reset the high water mark for NSS if range 0 value > 4096
						sysctl -w dev.nss.n2hcfg.n2h_high_water_core0=67392 >/dev/null 2>/dev/null
						#initially after init 4k buf for 5G and 4k for 2G will be allocated, later range will be configured
						sysctl -w dev.nss.n2hcfg.n2h_wifi_pool_buf=40960 >/dev/null 2>/dev/null
                        			update_ini_for_hk_dbs QCA8074V2_i.ini
					fi
					case "$hwmode" in
					*axa | *ac)
						if [ $soc_version_major = 2 ];then
							device_if "$phy" fc_buf0_max 32768
							device_if "$phy" fc_buf1_max 32768
							device_if "$phy" fc_buf2_max 32768
							device_if "$phy" fc_buf3_max 32768
						else
							device_if "$phy" fc_buf0_max 8192
							device_if "$phy" fc_buf1_max 8192
							device_if "$phy" fc_buf2_max 12288
							device_if "$phy" fc_buf3_max 32768
						fi
					;;
					*)
						if [ $soc_version_major = 2 ];then
							device_if "$phy" fc_buf0_max 16384
							device_if "$phy" fc_buf1_max 16384
							device_if "$phy" fc_buf2_max 16384
							device_if "$phy" fc_buf3_max 32768
						else
							device_if "$phy" fc_buf0_max 4096
							device_if "$phy" fc_buf1_max 8192
							device_if "$phy" fc_buf2_max 12288
							device_if "$phy" fc_buf3_max 16384
						fi
					;;
					esac
				fi
			;;
			*)
				case "$hwmode" in
				*ng)
					device_if "$phy" fc_buf0_max 5120
					device_if "$phy" fc_buf1_max 8192
					device_if "$phy" fc_buf2_max 12288
					device_if "$phy" fc_buf3_max 16384
					;;
				*ac)
					device_if "$phy" fc_buf0_max 8192
					device_if "$phy" fc_buf1_max 16384
					device_if "$phy" fc_buf2_max 24576
					device_if "$phy" fc_buf3_max 32768
					;;
				*)
					device_if "$phy" fc_buf0_max 5120
					device_if "$phy" fc_buf1_max 8192
					device_if "$phy" fc_buf2_max 12288
					device_if "$phy" fc_buf3_max 16384
					;;
				esac
				;;
			esac
		fi
	else
		local mp_512="$(ls /proc/device-tree/ | grep -rw "MP_512")"
		config_get hwmode "$device" hwmode auto
		if [ "$mp_512" == "MP_512" ]; then
			case "$hwmode" in
				*axa | *ac)
					# For 128 clients
					device_if "$phy" fc_buf_max 8192
					;;
				esac
		fi
	fi

        if [ $nss_wifi_olcfg == 0 ]; then
                sysctl -w dev.nss.n2hcfg.n2h_queue_limit_core0=2048 >/dev/null 2>/dev/null
                sysctl -w dev.nss.n2hcfg.n2h_queue_limit_core1=2048 >/dev/null 2>/dev/null
        else
                sysctl -w dev.nss.n2hcfg.n2h_queue_limit_core0=256 >/dev/null 2>/dev/null
                sysctl -w dev.nss.n2hcfg.n2h_queue_limit_core1=256 >/dev/null 2>/dev/null
        fi

        config_tx_fc_buf "$phy"

	# Advanced QCA wifi per-radio parameters configuration
	#jacky: remove chain mask setting for throughput issue, let driver auto detect
	#config_get txchainmask "$wifi_dev" txchainmask
	#[ -n "$txchainmask" ] && device_if "$phy" txchainmask "$txchainmask"

	#config_get rxchainmask "$wifi_dev" rxchainmask
	#[ -n "$rxchainmask" ] && device_if "$phy" rxchainmask "$rxchainmask"

        config_get_bool bcnburst "$wifi_dev" bcnburst 0
        [ "$bcnburst" -gt 0 ] && device_if "$phy" set_bcnburst "$bcnburst"

	case "$country" in  #for adaptive test COT issue
	40|56|100|191|196|203|208|233|246|250|276|300|348|372|380|442|807|499|528|578|616|620|642|688|703|705|724|752|756|792|804|826)
	    device_if $device aggr_burst 0 0 
	;;
	esac

        device_if "$phy" burst 1 #  INTEL IOT issue 
        device_if "$phy" dcs_enable 0 # adaptive issue
	#device_if "$phy" set_nohostcrash 1 #set it to avoid target assertion failed kernel panic, we will do wifi reload in other flow

        #SPF8.0 ADDED

        device_if "$device" enable_ol_stats 1

	if [ "$(uci -q get wifiprofile.snWifiConf.SUPPORT_ACS_ENHANCE_RANDOMIZE_SELECTION)" == "1" ]; then
		device_if "$device" acs_2g_allch 1
	fi
}
